/*	hardsync.c

	A Csound opcode library implementing ugens
	for hard sync.
	
	Anthony Kozar
	November 29, 2004

	Code for syncoscili borrowed from ugens2.c in canonical Csound

	Copyright (C) 1991,2004 Barry Vercoe, John ffitch,
	            Robin Whittle, and Anthony Kozar

	The hardsync opcode library is free software; you can redistribute it
	and/or modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	Csound is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with Csound; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
	02111-1307 USA
 */

#include	"hardsync.h"
#include	<math.h>

extern void	SyncOscInit(SYNC_OSC_DS* p);
extern void	SyncOscKK(SYNC_OSC_DS* p);

void	SyncSlaveInit(SYNC_SLAVE_DS* data)
{
	data->phase = FL(0.0);
	data->masterlast = FL(0.0);
	return;
}

void	SyncSlave(SYNC_SLAVE_DS* data)
{
	MYFLT		phaseincr, amp, last, cur;
	int		wasnegative;
	int		nsmps = ksmps;
	MYFLT*	out = data->aout;
	MYFLT*	master = data->amaster;
	
	phaseincr = *data->kfreq / esr;
	amp = *data->kamp;
	last = data->masterlast;
	wasnegative = (last < FL(0.0));
	do	{
		cur = *master++;
		/* reset phase if master signal changes from negative to nonnegative */
		if	(cur < FL(0.0))	wasnegative = true;
		else	{
			if	(wasnegative)	data->phase = FL(0.0);
			wasnegative = false;
		}
		last = cur;
		/* saw rises from 0.0 to 1.0, then -1.0 to 0.0 */
		if	(data->phase < FL(0.5))
			*out++ = amp * (FL(2.0) * data->phase);
		else	*out++ = amp * (FL(2.0) * (data->phase - FL(1.0)));
		data->phase = (MYFLT)fmod( data->phase + phaseincr, 1.0 );
	}
	while	(--nsmps);
	
	data->masterlast = last;
	return;
}


void	SyncMasterInit(SYNC_MASTER_DS* data)
{
	data->phase = FL(0.0);
	return;
}

void	SyncMaster(SYNC_MASTER_DS* data)
{
	MYFLT		phaseincr;
	int		nsmps = ksmps;
	MYFLT*	out = data->aout;
	
	phaseincr = *data->kfreq / esr;
	do	{
		/* saw rises from 0.0 to 1.0, then -1.0 to 0.0 */
		if	(data->phase < FL(0.5))
			*out++ = (FL(2.0) * data->phase);
		else	*out++ = (FL(2.0) * (data->phase - FL(1.0)));
		data->phase = (MYFLT)fmod( data->phase + phaseincr, 1.0 );
	}
	while	(--nsmps);
	
	return;
}


void	SyncPairInit(SYNC_PAIR_DS* data)
{
	data->phase = FL(0.0);
	data->masterlast = FL(0.0);
	data->masterphase = FL(0.0);
	return;
}

void	SyncPair(SYNC_PAIR_DS* data)
{
	MYFLT		phaseincr, masterphaseincr, amp, last, cur;
	int		wasnegative;
	int		nsmps = ksmps;
	MYFLT*	out = data->aout;
	
	phaseincr = *data->kfreq / esr;
	masterphaseincr = *data->kmfreq / esr;
	amp = *data->kamp;
	last = data->masterlast;
	wasnegative = (last < FL(0.0));
	do	{
		/* calculate master saw oscillator */
		if	(data->masterphase < FL(0.5))
			cur = (FL(2.0) * data->masterphase);
		else	cur = (FL(2.0) * (data->masterphase - FL(1.0)));
		data->masterphase = (MYFLT)fmod( data->masterphase + masterphaseincr, 1.0 );
		
		/* reset phase if master signal changes from negative to nonnegative */
		if	(cur < FL(0.0))	wasnegative = true;
		else	{
			if	(wasnegative)	data->phase = FL(0.0);
			wasnegative = false;
		}
		last = cur;
		/* saw rises from 0.0 to 1.0, then -1.0 to 0.0 */
		if	(data->phase < FL(0.5))
			*out++ = amp * (FL(2.0) * data->phase);
		else	*out++ = amp * (FL(2.0) * (data->phase - FL(1.0)));
		data->phase = (MYFLT)fmod( data->phase + phaseincr, 1.0 );
	}
	while	(--nsmps);
	
	data->masterlast = last;
	return;
}


void	SyncOscInit(SYNC_OSC_DS* p)
{
    FUNC        *ftp;

    if ((ftp = ftfind(p->ifn)) != NULL) {
      p->ftp = ftp;
      if (*p->iphs >= 0)
        p->lphs = ((long)(*p->iphs * FMAXLEN)) & PHMASK;
    }
    p->synclast = FL(0.0);
    
    return;
}

void	SyncOscKK(SYNC_OSC_DS* p)
{
    FUNC    *ftp;
    MYFLT   fract, v1, amp, *ar, *ftab;
    long    phs, inc, lobits;
    MYFLT	*sync, last, cur;		/* current and last values of sync input */
    int	wasnegative;      /* last value of sync input was negative if true */
    int     nsmps = ksmps;

    ftp = p->ftp;
    if (ftp==NULL) {
      perferror(Str(X_1108,"oscili: not initialised"));
      return;
    }
    lobits = ftp->lobits;
    phs = p->lphs;
    inc = (long)(*p->xcps * sicvt);
    amp = *p->xamp;
    ar = p->sr;
    sync = p->async;
    last = p->synclast;	  /* retrieve last sync value */
    wasnegative = (last < FL(0.0));
    do {
	cur = *sync++;
	/* reset phase if sync signal changes from negative to nonnegative */
	if	(cur < FL(0.0))	wasnegative = true;
	else	{
		if	(wasnegative)	phs = 0;
		wasnegative = false;
	}
	last = cur;

	/* calculate oscillator output */
      fract = PFRAC(phs);
      ftab = ftp->ftable + (phs >> lobits);
      v1 = *ftab++;
      *ar++ = (v1 + (*ftab - v1) * fract) * amp;
      phs += inc;
      phs &= PHMASK;
    }
    while (--nsmps);
    p->lphs = phs;
    p->synclast = last;
    return;
}


/* code for linking dynamic libraries under Csound 4 */

#define S       sizeof

static OENTRY localops[] = {
  { "syncslave", S(SYNC_SLAVE_DS), 5, "a", "akk", (SUBR)SyncSlaveInit, NULL, (SUBR)SyncSlave },
  { "syncmaster", S(SYNC_MASTER_DS), 5, "a", "k", (SUBR)SyncMasterInit, NULL, (SUBR)SyncMaster },
  { "syncpair", S(SYNC_PAIR_DS), 5, "a", "kkk", (SUBR)SyncPairInit, NULL, (SUBR)SyncPair },
  { "syncoscili", S(SYNC_OSC_DS), 5, "a", "akkio", (SUBR)SyncOscInit, NULL, (SUBR)SyncOscKK }
};

long opcode_size(void)
{
    return sizeof(localops);
}

OENTRY *opcode_init(GLOBALS *xx)
{
    pcglob = xx;
    return localops;
}
