/*	shape.c

	A Csound opcode library implementing ugens
	for various waveshaping methods.
	
	Anthony Kozar
	January 5, 2004
 */

#include "csdl.h"
#include <math.h>


typedef struct {
	OPDS	h;
	MYFLT	*aout, *ain, *kphaseadjust, *ifullscale;
	MYFLT	lastin, maxamplitude;
} PHASINE;

void	PhasineInit(PHASINE* data)
{
	data->lastin = FL(0.0);
	data->maxamplitude = *data->ifullscale;
	return;
}

void	Phasine(PHASINE* data)
{
	MYFLT		last, cur, phase, adjust, maxampl;
	int		nsmps = ksmps;
	MYFLT*	out = data->aout;
	MYFLT*	in = data->ain;
	
	adjust = *(data->kphaseadjust);
	last = data->lastin;
	maxampl = data->maxamplitude;
	do	{
		cur = *in++;
		phase = (MYFLT)asin(cur/maxampl) / (MYFLT)PI;	/* current "phase" value */
		if	(last < cur)	{
			if	(phase < 0)	phase = (FL(-1.0) - phase);
			else			phase = (FL(1.0) - phase);
		}
		last = cur;
		
		phase += adjust;						/* new phase value */
		/* wrap phase if using a lookup table */
		*out++ = (MYFLT)sin(phase*(MYFLT)PI) * maxampl;
	}
	while	(--nsmps);
	
	data->lastin = last;
	return;
	
}

typedef struct {
	OPDS	h;
	MYFLT	*aout, *ain, *kshapeamount, *ifullscale;
	MYFLT	maxamplitude;
} POWER_SHAPE;


void	PowerShapeInit(POWER_SHAPE* data)
{
	data->maxamplitude = *data->ifullscale;
	return;
}

void	PowerShape(POWER_SHAPE* data)
{
	MYFLT		cur, amt, maxampl;
	int		nsmps = ksmps;
	MYFLT*	out = data->aout;
	MYFLT*	in = data->ain;
	
	amt = *(data->kshapeamount);
	maxampl = data->maxamplitude;
	if	(amt == FL(0.0))	{					/* treat zero-power with care */
		do	{
			cur = *in++;
			if	(cur == FL(0.0))	*out++ = FL(0.0);	/* make 0^0 = 0 for continuity */
			else	*out++ = maxampl;				/* otherwise, x^0 = 1.0 */
		}
		while	(--nsmps);
	}
	else	{
		do	{
			cur = *in++ / maxampl;
			if	(cur < FL(0.0))				/* treat negatives with care */
				/* make output negative to avoid DC offsets and preserve continuity with even powers */
				*out++ = - (MYFLT)pow(-cur,amt) * maxampl;
			else	*out++ = (MYFLT)pow(cur,amt) * maxampl;
		}
		while	(--nsmps);
	}
	return;	
}


/* Phase distortion opcodes */

typedef struct {
	OPDS	h;
	MYFLT	*aout, *ain, *kwidth, *kcenter, *ifullscale;
} PD_CLIP;

void	PDClipInit(PD_CLIP* data)
{
	return;
}

void	PDClip(PD_CLIP* data)
{
	MYFLT		cur, low, high, maxampl, width, unwidth, center, outscalar;
	int		nsmps = ksmps;
	MYFLT*	out = data->aout;
	MYFLT*	in = data->ain;

	maxampl = *(data->ifullscale);
	width = (*(data->kwidth) > FL(1.0) ? FL(1.0) : 
		  (*(data->kwidth) < FL(0.0) ? FL(0.0) : *(data->kwidth)));
	unwidth = FL(1.0) - width;			/* width of unclipped region */
	
	/* the unclipped region can only shift left or right by up to width */
	if		(*(data->kcenter) < - width)	center = - width * maxampl;
	else if	(*(data->kcenter) > width)	center = width * maxampl;
	else							center = *(data->kcenter) * maxampl;
	
	low = center - unwidth*maxampl;		/* min value of unclipped input */
	high = unwidth*maxampl + center;		/* max value of unclipped input */
	outscalar = (unwidth == FL(0.0)) ? FL(0.0) : (FL(1.0) / unwidth);
	
	do	{
		cur = *in++;
		*out++ = (cur < low ? -maxampl : 
		         (cur > high ? maxampl : (outscalar * (cur-center))));
	}
	while (--nsmps);
	
	return;
}

typedef struct {
	OPDS	h;
	MYFLT	*aout, *ain, *kamount, *ifullscale, *iunipolar;
} PD_HALF;

/* Casio-style phase distortion with "pivot point" on the X axis */
void	PDHalfX(PD_HALF* data)
{
	MYFLT		cur, maxampl, midpoint, leftslope, rightslope;
	int		nsmps = ksmps;
	MYFLT*	out = data->aout;
	MYFLT*	in = data->ain;

	maxampl = *(data->ifullscale);
	midpoint = (*(data->kamount) > FL(1.0) ? maxampl : 
		     (*(data->kamount) < FL(-1.0) ? -maxampl : 
		     (*(data->kamount) * maxampl)));
	
	rightslope = maxampl / (maxampl - midpoint);
	leftslope  = maxampl / (midpoint + maxampl);
	
	do	{
		cur = *in++;
		if	(cur < midpoint)	*out++ = leftslope*(cur - midpoint);
		else				*out++ = rightslope*(cur - midpoint);
	}
	while (--nsmps);
	
	return;
}

/* Casio-style phase distortion with "pivot point" on the Y axis */
void	PDHalfY(PD_HALF* data)
{
	MYFLT		cur, maxampl, midpoint, leftslope, rightslope;
	int		nsmps = ksmps;
	MYFLT*	out = data->aout;
	MYFLT*	in = data->ain;

	maxampl = *(data->ifullscale);
	midpoint = (*(data->kamount) > FL(1.0) ? maxampl : 
		     (*(data->kamount) < FL(-1.0) ? -maxampl : 
		     (*(data->kamount) * maxampl)));
	
	rightslope = (maxampl - midpoint) / maxampl;
	leftslope  = (midpoint + maxampl) / maxampl;
	
	do	{
		cur = *in++;
		if	(cur < FL(0.0))	*out++ = leftslope*cur + midpoint;
		else				*out++ = rightslope*cur + midpoint;
	}
	while (--nsmps);
	
	return;
}


/* code for linking dynamic libraries under Csound 4 */

#define S       sizeof

static OENTRY localops[] = {
  { "phasine", S(PHASINE), 5, "a", "akp", (SUBR)PhasineInit, NULL, (SUBR)Phasine },
  { "powershape", S(POWER_SHAPE), 5, "a", "akp", (SUBR)PowerShapeInit, NULL, (SUBR)PowerShape },
  { "pdclip", S(PD_CLIP), 4, "a", "akkp", NULL, NULL, (SUBR)PDClip },
  { "pdhalf", S(PD_CLIP), 4, "a", "akp", NULL, NULL, (SUBR)PDHalfX },
  { "pdhalfy", S(PD_CLIP), 4, "a", "akp", NULL, NULL, (SUBR)PDHalfY }
};

long opcode_size(void)
{
    return sizeof(localops);
}

OENTRY *opcode_init(GLOBALS *xx)
{
    pcglob = xx;
    return localops;
}
